#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Operand
from .. import opcodes as OperandDef
from ..serialize import ValueType, AnyField, ListField


class Concatenate(Operand):
    _op_type_ = OperandDef.CONCATENATE

    _inputs = ListField('inputs', ValueType.key)
    _axis = AnyField('axis')

    @property
    def axis(self):
        return getattr(self, '_axis', None)
